﻿// BzrSharp
// Copyright (C) 2010  Carlos Flores <cafg10@gmail.com>

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using Bazaar.Commands;
using Bazaar.Results;
using CookComputing.XmlRpc;
using StructureMap;

namespace Bazaar
{
    public sealed class VersionControl : IDisposable
    {
        /// <summary>
        /// TPC/IP Port that will be used to run the server, this must match the client port
        /// </summary>
        private int port = 11113;
        private Process process;

        private Queue<Command> pendingCommands = new Queue<Command>();

        public VersionControl()
        {
            Start();
        }

        ~VersionControl()
        {
            Stop();
        }

        #region Process Start Up, Monitoring and Dettaching

        /// <summary>
        /// Initializes the bzr XMLRPC server for listening to <see cref="Commands"/>s
        /// </summary>
        public void Start()
        {
            process = new Process();
            process.StartInfo.UseShellExecute = false;
            process.StartInfo.RedirectStandardOutput = true;
            process.StartInfo.RedirectStandardError = true;
            process.StartInfo.CreateNoWindow = true;
            process.StartInfo.FileName = "bzr.exe";
            process.StartInfo.Arguments = "start-xmlrpc --verbose --port=" + port;
            process.Start();

            process.Exited += RestartProcess;
        }

        /// <summary>
        /// Restarts the process in case the XMLRPC server stops suddenly
        /// </summary>
        /// <param name="o">Calling object</param>
        /// <param name="args">Calling parameters</param>
        private void RestartProcess(object o, EventArgs args)
        {
            Start();
            ObjectFactory.EjectAllInstancesOf<IClient>();
            ObjectFactory.Configure(x =>
                {
                    x.ForSingletonOf<IClient>().Use(XmlRpcProxyGen.Create<IClient>());
                });
        }

        /// <summary>
        /// Stops the XMLRPC server by forcefully calling quit until the process exits suddenly
        /// or in a normal way. Warning: It will close any other bzr xmlrpc server running on the same port
        /// </summary>
        private void Stop()
        {
            process.Exited -= RestartProcess;

            RunPendingCommands();

            if (process.HasExited)
            {
                return;
            }

            IClient client = ObjectFactory.GetInstance<IClient>();
            client.Quit();

            try
            {
                while (!process.HasExited)
                {
                    client.Quit();
                }
            }
            catch (WebException)
            {
            }
        }

        public void Dispose()
        {
            Stop();
            process.Dispose();
            GC.SuppressFinalize(this);
        }

        #endregion Process Start Up, Monitoring and Dettaching

        #region Command Management, Queueing and Execution

        /// <summary>
        /// Checks if there are still commands waiting for execution
        /// </summary>
        public bool HasPendingCommands
        {
            get
            {
                return !(pendingCommands.Count == 0);
            }
        }

        /// <summary>
        /// Executes the command at the top of the queue
        /// </summary>
        /// <returns>The result of runing the <see cref="Command"/></returns>
        public IEnumerable<IResult> RunNextCommand()
        {
            Command command = pendingCommands.Dequeue();
            command.Run();
            return command.Results;
        }

        /// <summary>
        /// Adds a <see cref="Command"/> to the waiting queue
        /// </summary>
        /// <param name="command">The <see cref="Command"/> to enqueue</param>
        public void EnqueueCommand(Command command)
        {
            pendingCommands.Enqueue(command);
        }

        /// <summary>
        /// Executes all <see cref="Command"/>s that are still waiting
        /// </summary>
        /// <returns>The results of all <see cref="Command"/>s</returns>
        public IEnumerable<IResult> RunPendingCommands()
        {
            IList<IResult> results = new List<IResult>();

            while (HasPendingCommands)
            {
                foreach (IResult result in RunNextCommand())
                {
                    results.Add(result);
                }
            }

            return results;
        }

        #endregion Command Management, Queueing and Execution

        #region Utility functions

        public bool Ready
        {
            get
            {
                return !process.HasExited;
            }
        }

        public static IEnumerable<string> ListMethods()
        {
            var client = ObjectFactory.GetInstance<IClient>();

            return client.ListMethods();
        }

        public string RootFor(string file)
        {
            Root command = new Root(new Uri(file));
            command.Run();
            return command.Location;
        }

        #endregion Utility functions

        #region Version Control Functions

        public void Add(IEnumerable<string> files)
        {
            string root = RootFor(files.First());
            EnqueueCommand(new Add(root, files));
        }

        public void Remove(IEnumerable<string> files)
        {
            string root = RootFor(files.First());
            EnqueueCommand(new Remove(root, files));
        }

        public void Rename(string source, string destination)
        {
            string root = RootFor(source);
            EnqueueCommand(new Move(root, source, destination));
        }

        public WorkingTree GetWorkingTree(string source, IEnumerable<string> files)
        {
            string root = RootFor(source);
            Status status = new Status(root, files);
            LS ls = new LS(root);
            status.Run();
            ls.Run();

            return Parser.FullWorkingTree(ls.Response.Data, status.Response.Data);
        }

        #endregion Version Control Functions
    }
}
