﻿// BzrSharp
// Copyright (C) 2010  Carlos Flores <cafg10@gmail.com>

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Bazaar.Results;
using StructureMap;

namespace Bazaar.Commands
{
    /// <summary>
    /// Base class for providing command execution
    /// </summary>
    public abstract class Command
    {
        protected string name;
        protected IList<CommandOption> options = new List<CommandOption>();
        protected bool checkExitValue;
        protected Uri workingDirectory;
        protected Response response;

        protected Command()
        {
            workingDirectory = new Uri(Environment.CurrentDirectory);
        }

        protected Command(string directory)
        {
            workingDirectory = new Uri(directory);
        }

        protected Command(DirectoryInfo directory)
        {
            workingDirectory = new Uri(directory.FullName);
        }

        public Command(Uri workingDirectory)
        {
            this.workingDirectory = workingDirectory;
        }

        public virtual IEnumerable<IResult> Run()
        {
            var client = ObjectFactory.GetInstance<IClient>();
            string workingpath = workingDirectory.LocalPath.Replace("\n", "");
            response = new Response(client.Run(Arguments.ToArray(), workingpath));
            return Results;
        }

        /// <summary>
        /// Creates a list of arguments to be used by the command, including the name of the bzr executable
        /// </summary>
        public virtual IList<string> Arguments
        {
            get
            {
                IList<string> arguments = new List<string>();
                arguments.Add("bzr");
                arguments.Add(name);
                string options = OptionsToString();
                if (!String.IsNullOrEmpty(options))
                {
                    arguments.Add(OptionsToString());
                }
                return arguments;
            }
        }

        /// <summary>
        /// Helps retrieving result messages from the command, MUST be overrided by
        /// Inheriting classes if they give extra important info
        /// </summary>
        public virtual IEnumerable<IResult> Results
        {
            get
            {
                IList<IResult> result = new List<IResult>();
                result.Add(new StringResult(response.Data));
                return result;
            }
        }

        #region Convenience Methods

        public string OptionsToString()
        {
            string commandLine = String.Empty;
            foreach (CommandOption option in Options)
            {
                if (option == null)
                {
                    continue;
                }
                string str = option.ToString();
                if (!String.IsNullOrEmpty(str))
                {
                    commandLine += str;
                }
            }
            return commandLine;
        }

        public override string ToString()
        {
            return String.Concat(Arguments.ToArray());
        }

        #endregion Convenience Methods

        #region Accesor Properties

        /// <summary>
        /// Access the list of <see cref="Options"/> that will be used with the command
        /// </summary>
        public virtual IList<CommandOption> Options
        {
            get { return options; }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public bool CheckExitValue
        {
            get { return checkExitValue; }
            set { checkExitValue = value; }
        }

        public Uri WorkingDirectory
        {
            get { return workingDirectory; }
            set { workingDirectory = value; }
        }

        public Response Response
        {
            get { return response; }
            protected set { response = value; }
        }

        #endregion Accesor Properties
    }
}
