// BzrSharp
// Copyright (C) 2010  Carlos Flores <cafg10@gmail.com>

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using Bazaar.Results;

namespace Bazaar
{
    /// <summary>
    /// Contains many methods to parse the responses from the bazaar XMLRPC server
    /// </summary>
    internal static class Parser
    {
        /// <summary>
        /// Converts an xmlls response in a <see cref="Item"/> collection
        /// </summary>
        /// <param name="stream">The data of the xmlls response</param>
        /// <returns>The parsed <see cref="Item"/>s</returns>
        public static IEnumerable<Item> LS(string stream)
        {
            XDocument doc = XDocument.Parse(stream);
            IEnumerable<Item> items = (from item in doc.Descendants("item")
                                       select new Results.Item
                                       {
                                           Kind = (Kind)Enum.Parse(typeof(Results.Kind), item.Element("kind").Value, true),
                                           FileName = item.Element("path").Value,
                                           Status = (SourceControl)Enum.Parse(typeof(SourceControl),
                                                                           item.Element("status_kind").Value.Replace(" ", ""),
                                                                           true),
                                           Id = (string)item.Element("id"),
                                       });

            return items;
        }

        /// <summary>
        /// Creates a complete <see cref="WorkingTree"/> from a xmlstatus and xmlls responses
        /// </summary>
        /// <param name="xmlls">The data of the xmlls respone</param>
        /// <param name="xmlstatus">The data of the xmlstatus respone</param>
        /// <returns>A complete <see cref="WorkingTree"/></returns>
        public static WorkingTree FullWorkingTree(string xmlls, string xmlstatus)
        {
            WorkingTree workingTree = EmptyWorkingTree(xmlstatus);

            workingTree.Changes.Union(Status(xmlstatus).ToDictionary(x => x.Code, x => x));
            workingTree.Items.Union(LS(xmlls));

            return workingTree;
        }

        /// <summary>
        /// Converts a xmlstatus response in a <see cref="WorkingTree"/> description
        /// </summary>
        /// <param name="stream">The data of the xmlstatus respone</param>
        /// <returns>An incomplete <see cref="WorkingTree"/></returns>
        public static WorkingTree EmptyWorkingTree(string stream)
        {
            XDocument doc = XDocument.Parse(stream);

            WorkingTree workingTree = new WorkingTree(doc.Root.Attribute("workingtree_root").Value);

            return workingTree;
        }

        /// <summary>
        /// Converts a xmlstatus response into a <see cref="Results.FileStatus"/> collection
        /// </summary>
        /// <param name="stream">The data of the xmlstatus response</param>
        /// <returns>The parsed <see cref="Results.FileStatus" />es</returns>
        public static IEnumerable<FileStatus> Status(string stream)
        {
            XDocument doc = XDocument.Parse(stream);

            DirectoryInfo directory = new DirectoryInfo((from s in doc.Descendants("status")
                                                         select s.Attribute("workingtree_root").Value).First());

            return from element in doc.Root.Elements() select Status(element, directory);
        }

        /// <summary>
        /// Creates a <see cref="FileStatus"/> from an <see cref="XElement"/> given a base directory
        /// </summary>
        /// <param name="element">The <see cref="XElement"/> to parse</param>
        /// <param name="directory">The directory that serves as root for the <see cref="WorkingTree"/></param>
        /// <returns>The created <see cref="FileStatus"/></returns>
        public static FileStatus Status(XElement element, DirectoryInfo directory)
        {
            return new FileStatus
            {
                Code = (StatusCode)Enum.Parse(typeof(StatusCode),
                                                    element.Name.LocalName.Replace("-", ""),
                                                    true),
                FileName = element.Value,
                Root = directory,
            };
        }

        /// <summary>
        /// Converts an <see cref="XElement"/> into a <see cref="FileStatus"/>
        /// </summary>
        /// <param name="element">The <see cref="XElement"/> to parse</param>
        /// <returns>The <see cref="FileStatus"/></returns>
        public static FileStatus Status(XElement element)
        {
            return new FileStatus
            {
                Code = (StatusCode)Enum.Parse(typeof(StatusCode),
                                                    element.Name.LocalName.Replace("-", ""),
                                                    true),
                FileName = element.Value,
            };
        }

        /// <summary>
        /// Parses xmllog Response from a Log
        /// </summary>
        /// <param name="stream">The data of the xmllog response</param>
        /// <returns>A collection of <see cref="LogMessage"/>s</returns>
        public static IEnumerable<LogMessage> Log(string stream)
        {
            XDocument doc = XDocument.Parse(stream);

            return from element in doc.Root.Elements("log") select Log(element);
        }

        /// <summary>
        /// Creates a <see cref="LogMessage"/> from an <see cref="XElement"/>
        /// </summary>
        /// <param name="element">The <see cref="XElement"/> to parse</param>
        /// <returns>A new <see cref="LogMessage"/></returns>
        public static LogMessage Log(XElement element)
        {
            LogMessage logMessage = new LogMessage
            {
                Revision = int.Parse(element.Element("revno").Value, CultureInfo.InvariantCulture),
                Committer = element.Element("committer").Value,
                Nick = (string)element.Element("branch-nick"),
                Message = element.Element("message").Value,
                Timestamp = element.Element("timestamp").Value,
            };

            logMessage.Files.Union(from files in element.Elements("affected-files")
                                   from status in files.Elements()
                                   select Status(status));

            logMessage.Merged.Union(from merges in element.Elements("merge")
                                    from log in merges.Elements("log")
                                    select Log(log)
                                    );

            return logMessage;
        }

        /// <summary>
        /// Creates an <see cref="Annotation"/> from a string
        /// </summary>
        /// <param name="stream">The string to parse</param>
        /// <returns>A new <see cref="Annotation"/></returns>
        public static Annotation Annotate(string stream)
        {
            XDocument doc = XDocument.Parse(stream);
            Annotation annotation = new Annotation
                                    {
                                        Root = new DirectoryInfo(doc.Root.Attribute("workingtree_root").Value)
                                    };

            annotation.Entries.Union(from element in doc.Elements()
                                     select new Entry
                                         {
                                             Author = element.Attribute("author").Value,
                                             Revision = int.Parse(element.Attribute("revno").Value, CultureInfo.InvariantCulture),
                                             Date = DateTime.Parse(element.Attribute("date").Value, CultureInfo.InvariantCulture),
                                         });

            return annotation;
        }
    }
}
