﻿// BzrSharp
// Copyright (C) 2010  Carlos Flores <cafg10@gmail.com>

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Bazaar.Commands;
using Bazaar.Results;
using NUnit.Framework;
using StructureMap;

namespace Bazaar.Tests
{
    [TestFixture]
    public class Main
    {
        private VersionControl bzr;
        private DirectoryInfo temp;
        private FileInfo tempfile;
        private DirectoryInfo branch;

        [SetUp]
        public void SetUp()
        {
            BazaarBootstrapper.Bootstrap();

            bzr = ObjectFactory.GetInstance<VersionControl>();
            temp = new DirectoryInfo(Path.Combine(Path.GetTempPath(), Path.GetRandomFileName()));
            branch = new DirectoryInfo(Path.Combine(Path.GetTempPath(), Path.GetRandomFileName()));
            tempfile = new FileInfo(Path.Combine(temp.FullName, Path.GetRandomFileName()));
            temp.Create();
        }

        [Test]
        public void QueryServer()
        {
            Assert.IsNotEmpty(VersionControl.ListMethods().ToArray());
        }

        [Test]
        public void EnqueueCommand()
        {
            bzr.EnqueueCommand(new WhoAmI());
            bzr.RunNextCommand();
        }

        [Test]
        public void InitBranch()
        {
            bzr.EnqueueCommand(new Init(temp));
            bzr.RunNextCommand();
            Assert.True(Directory.Exists(Path.Combine(temp.FullName, ".bzr")));
        }

        [Test]
        public void InitSpacedBranch()
        {
            DirectoryInfo spaced = new DirectoryInfo(Path.Combine(temp.FullName, "Spaced Directory"));
            spaced.Create();
            bzr.EnqueueCommand(new Init(spaced));
        }

        [Test]
        public void BranchRemote()
        {
            bzr.EnqueueCommand(new Branch("lp:bazaarsharp", branch));
            bzr.RunNextCommand();
            Assert.False(branch.GetFileSystemInfos().ToList().Count == 0);
        }

        [Test]
        public void AddFile()
        {
            bzr.EnqueueCommand(new Init(temp));
            bzr.RunNextCommand();
            tempfile.Create();
            IList<FileSystemInfo> list = new List<FileSystemInfo>();
            list.Add(tempfile);
            bzr.EnqueueCommand(new Add(temp, list));
            bzr.RunNextCommand();
        }

        [Test]
        public void BranchActivity()
        {
            bzr.EnqueueCommand(new Init(temp));
            bzr.RunPendingCommands();
            tempfile.Create().Close();
            IList<string> list = new List<string>();
            list.Add(tempfile.FullName);
            bzr.Add(list);
            string destiny = Path.GetRandomFileName();
            FileInfo destination = new FileInfo(Path.Combine(temp.FullName, destiny));
            bzr.EnqueueCommand(new Move(temp, list, destination));
            bzr.EnqueueCommand(new Status(temp));
            bzr.EnqueueCommand(new Commit(temp, "Test Commit"));
            IEnumerable<IResult> results = bzr.RunPendingCommands();
            foreach (IResult result in results)
            {
                Console.WriteLine(result.ToString());
            }
        }

        [Test]
        public void WorkingTree()
        {
            bzr.EnqueueCommand(new Init(temp));
            bzr.RunPendingCommands();
            Status command = new Status(temp);
            bzr.EnqueueCommand(command);
            tempfile.Create().Close();
            IList<string> list = new List<string>();
            list.Add(tempfile.FullName);
            bzr.Add(list);
            bzr.RunPendingCommands();
            WorkingTree tree = bzr.GetWorkingTree(temp.FullName, list);
        }

        [TearDown]
        public void CleanUp()
        {
            bzr.Dispose();
            try
            {
                if (branch.Exists)
                {
                    branch.Delete(true);
                }
                temp.Delete(true);
                if (tempfile.Exists)
                {
                    tempfile.Delete();
                }
            }
            catch (IOException)
            {
            }
        }
    }
}
